from abc import ABC, abstractclassmethod

from tukaan._info import System
from tukaan._tcl import Tcl
from tukaan._utils import mac_only, windows_only


class Theme(ABC):
    @abstractclassmethod
    def _use(cls):
        pass


class ClamTheme(Theme):
    @classmethod
    def _use(cls):
        Tcl.call(None, "ttk::style", "theme", "use", "clam")


class Win32Theme(Theme):
    @classmethod
    @windows_only
    def _use(cls):
        Tcl.call(None, "ttk::style", "theme", "use", "vista")


class AquaTheme(Theme):
    @classmethod
    @mac_only
    def _use(cls):
        Tcl.call(None, "ttk::style", "theme", "use", "aqua")


def native_theme() -> Theme:
    if System.os == "Windows":
        return Win32Theme
    elif System.os == "macOS":
        return AquaTheme
    return ClamTheme



from __future__ import annotations

import collections
from typing import Callable

from ._events import EventMixin
from ._layout import ContainerGrid, Geometry, Grid, Position, ToplevelGrid
from ._mixins import DnDMixin, GeometryMixin, VisibilityMixin, WidgetMixin
from ._props import cget, config
from ._tcl import Tcl
from ._utils import _commands, _widgets, count
from ._wm import WindowManager
from .widgets.tooltip import ToolTipProvider


def generate_pathname(widget: TkWidget, parent: TkWidget) -> str:
    klass = widget.__class__
    count = next(parent._child_type_count[klass])

    return ".".join([parent._name, f"{klass.__name__.lower()}_{count}"])


class Container:
    ...


class InputControl:
    ...


class OutputDisplay:
    ...


class XScrollable:
    def x_scroll(self, *args) -> None:
        Tcl.call(None, self, "xview", *args)

    @property
    def on_xscroll(self) -> Callable:
        return _commands[cget(self, str, "-xscrollcommand")]

    @on_xscroll.setter
    def on_xscroll(self, value: Callable) -> None:
        config(self, xscrollcommand=value)


class YScrollable:
    def y_scroll(self, *args) -> None:
        Tcl.call(None, self, "yview", *args)

    @property
    def on_yscroll(self) -> Callable:
        return _commands[cget(self, str, "-yscrollcommand")]

    @on_yscroll.setter
    def on_yscroll(self, value: Callable) -> None:
        config(self, yscrollcommand=value)


class TkWidget(WidgetMixin, EventMixin, VisibilityMixin, DnDMixin):
    """Base class for every Tk widget"""

    _name: str
    _tcl_class: str

    def __init__(self):
        self._children = {}
        self._child_type_count = collections.defaultdict(lambda: count())

        _widgets[self._name] = self


class WindowBase(TkWidget, WindowManager, Container):
    def __init__(self):
        self.grid = ToplevelGrid(self)

        TkWidget.__init__(self)


class WidgetBase(TkWidget, GeometryMixin):
    def __init__(self, parent: TkWidget, tooltip: str | None = None, **kwargs):
        assert isinstance(parent, Container), "parent must be a container"

        self._name = self._lm_path = generate_pathname(self, parent)
        self.parent = parent
        self.parent._children[self._name] = self

        TkWidget.__init__(self)

        self.layout_manager = None
        self.grid = ContainerGrid(self) if isinstance(self, Container) else Grid(self)
        self.geometry = Geometry(self)
        self.position = Position(self)

        Tcl.call(None, self._tcl_class, self._name, *Tcl.to_tcl_args(**kwargs))

        if tooltip:
            ToolTipProvider.add(self, tooltip)

    def destroy(self):
        """Destroys this widget, and removes it from the screen"""

        Tcl.call(None, "destroy", self._name)

        del self.parent._children[self._name]
        del _widgets[self._name]

    @property
    def tooltip(self) -> str | None:
        return ToolTipProvider.get(self)

    @tooltip.setter
    def tooltip(self, value: str) -> None:
        ToolTipProvider.update(self, value)



from typing import Any, Callable, Optional

from ._tcl import Tcl


class Timeout:
    _after_id: str = ""
    _repeat: bool = False
    state: str = "not started"

    def __init__(self, seconds: float, target: Callable[[], Any], *, args=(), kwargs={}) -> None:
        assert callable(target), "target must be callable"

        self.seconds = seconds
        self.target = target

    def __repr__(self) -> str:
        name = self.target.__name__
        if name != "<lambda>":
            name += "()"

        return f"<{self.state.capitalize()} `{name}` timeout at {hex(id(self))}>"

    def run_once(self):
        try:
            self.target()
        except Exception as e:
            print(e)
            self.state = "failed"

    def run_now(self):
        self.cancel()
        self.run_once()

    def run(self):
        try:
            self.target()
            if self._repeat:
                return self.start()
        except Exception as e:
            print(e)
            self.state = "failed"
        else:
            self.state = "succesfully completed"

    __call__ = run

    def start(self) -> None:
        self._after_id = Tcl.call(str, "after", int(self.seconds * 1000), self.__call__)
        self.state = "pending"

    def repeat(self) -> None:
        self._repeat = True
        self.start()

    def cancel(self) -> None:
        if self.state != "pending":
            raise RuntimeError(f"cannot cancel a {self.state} timeout")

        command, _ = Tcl.call((str,), "after", "info", self._after_id)
        Tcl.call(None, "after", "cancel", command)
        Tcl.delete_cmd(command)

        self._repeat = False
        self.state = "cancelled"

    @property
    def is_repeated(self) -> bool:
        return self._repeat

    @is_repeated.setter
    def is_repeated(self, repeat: bool) -> None:
        self._repeat = repeat


class Timer:
    @staticmethod
    def schedule(seconds: float, target: Callable[[Any], Any], *, args=(), kwargs={}) -> None:
        Tcl.call(
            str, "after", int(seconds * 1000), Tcl.create_cmd(target, args=args, kwargs=kwargs)
        )

    @staticmethod
    def wait(seconds: float) -> None:
        script = f"""
        set tukaan_waitvar 0
        after {int(seconds * 1000)} {{set tukaan_waitvar 1}}
        tkwait variable tukaan_waitvar"""

        Tcl.eval(None, script)

    @staticmethod
    def delayed(seconds: float) -> Callable:
        def decorator(func: Callable) -> Callable:
            def wrapper(*args, **kwargs) -> Any:
                Timer.wait(seconds)
                return func(*args, **kwargs)

            return wrapper

        return decorator



from __future__ import annotations

from typing import Callable

from tukaan._base import InputControl, TkWidget, WidgetBase
from tukaan._props import cget, config, focusable, link, orientation, value
from tukaan._variables import Float, Integer
from tukaan.enums import Orientation


class Slider(WidgetBase, InputControl):
    _tcl_class = "ttk::scale"

    focusable = focusable
    link = link
    orientation = orientation
    value = value

    def __init__(
        self,
        parent: TkWidget,
        *args,
        focusable: bool | None = None,
        link: Integer | Float | None = None,
        max: float | None = None,
        min: float | None = None,
        on_move: Callable[[float], None] | None = None,
        orientation: Orientation | None = None,
        tooltip: str | None = None,
        value: int | None = None,
    ) -> None:

        if (max is not None and args) or (min is not None and len(args) == 2):
            raise TypeError(
                f"Slider() got multiple values for argument {('max' if max is not None else 'min')!r}"
            )

        if not args and max is None:
            max = 100
        elif len(args) == 1:
            [max] = args
        elif len(args) == 2:
            min, max, *_ = args

        self._original_cmd = on_move
        if on_move is not None:
            func = on_move
            on_move = lambda x: func(float(x))

        self._variable = link

        WidgetBase.__init__(
            self,
            parent,
            command=on_move,
            from_=min,
            orient=orientation,
            takefocus=focusable,
            to=max,
            tooltip=tooltip,
            value=value,
            variable=link,
        )

    def _repr_details(self):
        return f"min={self.min!r}, max={self.max!r}, value={self.value!r}"

    @property
    def bounds(self):
        return cget(self, float, "-from"), cget(self, float, "-to")

    @bounds.setter
    def bounds(self, value: tuple[float, float]):
        from_, to = value
        config(self, from_=from_)
        config(self, to=to)

    @property
    def min(self):
        return cget(self, float, "-from")

    @min.setter
    def min(self, value: float):
        config(self, from_=value)

    @property
    def max(self):
        return cget(self, float, "-to")

    @max.setter
    def max(self, value: float):
        config(self, to=value)

    @property
    def on_move(self) -> Callable[[float], None] | None:
        return self._original_cmd

    @on_move.setter
    def on_move(self, func: Callable[[float], None]) -> None:
        self._original_cmd = func
        value = lambda x: func(float(x))
        config(self, command=value)



from __future__ import annotations

from dataclasses import dataclass

from ._info import Screen
from .exceptions import ColorError


def mm(amount):
    return round(amount / (Screen.ppi / 25.4), 2)


def cm(amount):
    return round(amount / (Screen.ppi / 2.54), 2)


def inch(amount):
    return round(amount / Screen.ppi, 2)


@dataclass
class ScreenDistance:
    pixels: float

    def __init__(
        self,
        px: float | None = None,
        mm: float | None = None,
        cm: float | None = None,
        inch: float | None = None,
    ) -> None:
        self._ppi = ppi = Screen.ppi

        self.pixels = 0
        if px is not None:
            self.pixels += px
        if mm is not None:
            self.pixels += mm * (ppi * 25.4)
        if cm is not None:
            self.pixels += cm * (ppi * 2.54)
        if inch is not None:
            self.pixels += inch * ppi

    def __to_tcl__(self) -> str:
        return str(self.pixels)

    @classmethod
    def __from_tcl__(cls, tcl_value: str) -> ScreenDistance:
        unit = tcl_value[-1]
        value = int(tcl_value[:-1])

        if unit == "c":
            return cls(cm=value)
        if unit == "m":
            return cls(mm=value)
        if unit == "i":
            return cls(inch=value)

        return cls(px=value)

    def __eq__(self, other: ScreenDistance):
        if not isinstance(other, ScreenDistance):
            raise TypeError

        return self.pixels == other.pixels

    def __gt__(self, other: ScreenDistance):
        if not isinstance(other, ScreenDistance):
            raise TypeError

        return self.pixels > other.pixels

    def __lt__(self, other: ScreenDistance):
        if not isinstance(other, ScreenDistance):
            raise TypeError

        return self.pixels < other.pixels

    @property
    def px(self) -> float:
        return round(self.pixels, 2)

    @property
    def mm(self) -> float:
        return mm(self.pixels)

    @property
    def cm(self) -> float:
        return cm(self.pixels)

    @property
    def inch(self) -> float:
        return inch(self.pixels)



class TclError(Exception):
    ...


class AppError(Exception):
    ...


class WidgetError(Exception):
    ...


class ColorError(Exception):
    ...


class FontError(Exception):
    ...


class LayoutError(Exception):
    ...


class CellNotFoundError(LayoutError):
    ...



import sys
from pathlib import Path

from . import Serif
from ._base import WindowBase
from ._tcl import Tcl
from .exceptions import AppError


class App(WindowBase):
    _exists = False
    _name = _lm_path = ".app"
    _wm_path = "."

    def __init__(
        self,
        title: str = "Tukaan",
        *,
        width: int = 300,
        height: int = 200,
    ) -> None:
        if not App._exists:
            App._exists = True
        else:
            raise AppError("can't create multiple App instances. Use tukaan.Window instead.")

        super().__init__()

        Tcl.init()
        Tcl.eval(None, "pack [ttk::frame .app] -expand 1 -fill both")

        Tcl.call(None, "bind", self, "<Map>", self._generate_state_event)
        Tcl.call(None, "bind", self, "<Unmap>", self._generate_state_event)
        Tcl.call(None, "bind", self, "<Configure>", self._generate_state_event)
        Tcl.call(None, "wm", "protocol", self._wm_path, "WM_DELETE_WINDOW", self.destroy)

        self._set_title(title)
        self._set_size((width, height))

        self._init_tkdnd()
        Serif.load()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, _):
        if exc_type is None:
            return self.run()

        raise exc_type(exc_value) from None

    def _init_tkdnd(self):
        os = {"linux": "linux", "darwin": "mac", "win32": "win"}[sys.platform]
        os += "-x64" if sys.maxsize > 2**32 else "-x32"

        Tcl.call(None, "lappend", "auto_path", Path(__file__).parent / "tkdnd" / os)
        Tcl.call(None, "package", "require", "tkdnd")

    def destroy(self) -> None:
        """Quit the entire Tcl interpreter"""

        Tcl.call(None, "destroy", self._name)
        Tcl.call(None, "destroy", self._wm_path)

        Tcl.quit_interp()

    def run(self) -> None:
        """Starts the main event loop"""

        Tcl.main_loop()



from __future__ import annotations

from typing import TYPE_CHECKING

from tukaan._tcl import Tcl
from tukaan.exceptions import TclError

if TYPE_CHECKING:
    from tukaan._base import WidgetBase


class ToolTipProvider:
    _widgets: dict[str, str] = {}
    _schedule_cmd: str = ""
    _hide_cmd: str = ""
    _show_cmd: str = ""
    _setup_done = False
    _after_id: str = ""
    _can_show = True

    @classmethod
    def setup(cls) -> None:
        cls._schedule_cmd = Tcl.to(cls.schedule)
        cls._hide_cmd = Tcl.to(cls.hide)
        cls._show_cmd = Tcl.to(cls.show)

        try:
            Tcl.eval(None, "ttk::style layout ToolTip")
        except TclError:
            style = "TButton"
        else:
            style = "ToolTip"

        Tcl.call(None, "toplevel", ".tooltip")
        Tcl.call(None, "wm", "withdraw", ".tooltip")
        Tcl.eval(None, f"pack [ttk::label .tooltip.label -style {style}] -expand 1 -fill both")
        Tcl.call(None, "wm", "overrideredirect", ".tooltip", 1)

        cls._setup_done = True

    @classmethod
    def add(cls, owner: WidgetBase, message: str) -> None:
        if not cls._setup_done:
            cls.setup()

        cls._widgets[owner._lm_path] = message

        Tcl.call(None, "bind", owner._lm_path, "<Enter>", f"+ {cls._schedule_cmd} %W")
        Tcl.call(None, "bind", owner._lm_path, "<Leave>", f"+ {cls._hide_cmd}")
        Tcl.call(None, "bind", owner._lm_path, "<ButtonPress>", f"+ {cls._hide_cmd}")

    @classmethod
    def update(cls, owner: WidgetBase, message: str) -> None:
        if owner._lm_path not in cls._widgets:
            return cls.add(owner, message)

        cls._widgets[owner._lm_path] = message

    @classmethod
    def get(cls, owner: WidgetBase) -> str | None:
        return cls._widgets.get(owner._lm_path)

    @classmethod
    def schedule(cls, widget: str) -> None:
        cls._can_show = True

        message = cls._widgets.get(widget)
        if message is None:
            return

        Tcl.call(None, ".tooltip.label", "configure", "-text", message)
        cls._after_id = Tcl.call(str, "after", 1000, cls._show_cmd, widget)

    @classmethod
    def hide(cls) -> None:
        cls._can_show = False  # after cancel sometime doesn't work, so this is a hack

        Tcl.call(None, "after", "cancel", cls._show_cmd)
        Tcl.call(None, "after", "cancel", cls._hide_cmd)
        Tcl.call(None, "wm", "withdraw", ".tooltip")

    @classmethod
    def show(cls, widget: str) -> None:
        if not cls._can_show:
            return

        owner_x = Tcl.call(int, "winfo", "rootx", widget)
        owner_width = Tcl.call(int, "winfo", "reqwidth", widget)
        tip_width = Tcl.call(int, "winfo", "reqwidth", ".tooltip")
        tip_x = owner_x + owner_width // 2 - tip_width // 2

        owner_y = Tcl.call(int, "winfo", "rooty", widget)
        tip_height = Tcl.call(int, "winfo", "reqheight", ".tooltip")
        tip_y = owner_y - tip_height - 5

        if tip_y <= Tcl.eval(int, f"winfo rooty [winfo toplevel {widget}]"):
            tip_y = owner_y + Tcl.call(int, "winfo", "reqheight", widget) + 5

        Tcl.call(None, "wm", "geometry", ".tooltip", f"+{tip_x}+{tip_y}")
        Tcl.call(None, "wm", "deiconify", ".tooltip")

        Tcl.call(str, "after", 10000, cls._hide_cmd)



from __future__ import annotations

from ._tcl import Tcl
from ._utils import _variables, counts


class ControlVariable:
    _type_spec: type
    _default: float | str | bool

    def __init__(self, value=None, name=None):
        if name is None:
            name = f"tukaan_{self._type_spec.__name__}_var_{next(counts['variable'])}"

        if value is None:
            value = self._default

        _variables[name] = self
        self._name = name
        self.set(value)

    def __repr__(self):
        return f"tukaan.{type(self).__name__} control variable: tcl_name={self._name}, value={self.get()!r}"

    __str__ = __repr__

    def __hash__(self):
        return hash((self._type_spec, self._name))

    def __to_tcl__(self):
        return self._name

    @classmethod
    def __from_tcl__(cls, value):
        return _variables[value]

    def set(self, new_value):
        return Tcl.call(self._type_spec, "set", self._name, new_value)

    def get(self):
        return Tcl.call(self._type_spec, "set", self._name)

    @property
    def value(self):
        return self.get()

    @value.setter
    def value(self, value: float | str | bool):
        self.set(value)


class String(ControlVariable):
    _type_spec = str
    _default = ""


class Integer(ControlVariable):
    _type_spec = int
    _default = 0


class Float(ControlVariable):
    _type_spec = float
    _default = 0.0


class Boolean(ControlVariable):
    _type_spec = bool
    _default = False

    def __invert__(self) -> bool:
        inverted = not Tcl.call(bool, "set", self._name)
        Tcl.call(None, "set", self._name, inverted)
        return inverted



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from ._tcl import Tcl
from .enums import Align, Anchor
from .exceptions import LayoutError

if TYPE_CHECKING:
    from ._base import WidgetBase, WindowBase


class LayoutManager(ABC):
    _type: str

    def __init__(self, owner: WidgetBase) -> None:
        self._widget = owner

    @abstractmethod
    def __call__(self, *args, **kwargs) -> None:
        pass

    def _cget(self, return_type: type[Any], option: str) -> int | str | None:
        try:
            result = Tcl.call({option: return_type}, self._type, "info", self._widget)[option]
        except KeyError:
            return None
        else:
            return result

    def _config(self, **kwargs) -> None:
        Tcl.call(None, self._type, "configure", self._widget, *Tcl.to_tcl_args(**kwargs))


class Grid(LayoutManager):
    _type = "grid"

    def __call__(
        self,
        row: int | None = 0,
        col: int | None = 0,
        *,
        align: tuple[Align | None, Align | None] | None = None,
        cell: str | None = None,
        colspan: int | None = None,
        margin: int | tuple[int, ...] | None = None,
        rowspan: int | None = None,
    ) -> None:
        sticky = self._parse_align(align)
        padx, pady = self._parse_margin(margin)

        if cell:
            self._set_cell(cell)
            row = col = rowspan = colspan = None

        Tcl.call(
            None,
            "grid",
            "configure",
            self._widget._lm_path,
            *Tcl.to_tcl_args(
                row=row,
                column=col,
                rowspan=rowspan,
                columnspan=colspan,
                sticky=sticky,
                padx=padx,
                pady=pady,
            ),
        )

    def _set_cell(self, cell_name: str) -> None:
        try:
            cell = self._widget.parent.grid._cells_values[cell_name]
        except KeyError:
            raise LayoutError(f"cell {cell_name!r} doesn't exists") from None

        Tcl.call(
            None,
            "grid",
            "configure",
            self._widget,
            *Tcl.to_tcl_args(
                row=cell["row"],
                column=cell["col"],
                rowspan=cell["rowspan"],
                columnspan=cell["colspan"],
            ),
        )

        self._widget.parent.grid._cell_managed_children[self._widget] = cell_name

    def _parse_align(self, align: tuple[Align | None, Align | None] | None) -> str:
        if align is None or align == (None, None):
            return ""

        result = ""

        if isinstance(align, (tuple, list)) and len(align) == 1:
            align = align * 2
        elif isinstance(align, Align):
            align = (align,) * 2

        hor, vert = align

        if vert is Align.Start:
            result += "n"
        elif vert is Align.End:
            result += "s"
        elif vert is Align.Stretch:
            result += "ns"

        if hor is Align.Start:
            result += "w"
        elif hor is Align.End:
            result += "e"
        elif hor is Align.Stretch:
            result += "ew"

        return result

    def _parse_margin(
        self, margin: int | tuple[int, ...] | None
    ) -> tuple[tuple[int, int], tuple[int, int]] | tuple[None, None]:

        if isinstance(margin, int):
            return ((margin,) * 2,) * 2
        elif isinstance(margin, (tuple, list)):
            length = len(margin)

            if length == 2:
                vert, hor = margin
                return (hor,) * 2, (vert,) * 2
            elif length == 3:
                top, hor, bottom = margin
                return (hor,) * 2, (top, bottom)
            elif length == 4:
                top, right, bottom, left = margin
                return (left, right), (top, bottom)

        return None, None

    def _get_pad(self):
        result = Tcl.call(
            {
                "-padx": (int,),
                "-pady": (int,),
            },
            "grid",
            "info",
            self._widget,
        )

        return result["-padx"], result["-pady"]

    # Properties #

    @property
    def location(self) -> tuple[int | None, int | None]:
        return self._cget(int, "-row"), self._cget(int, "-column")

    @location.setter
    def location(self, value: tuple[int, int]) -> None:
        row, col = value
        self._config(row=row, col=col)

    @property
    def row(self) -> int | None:
        return self._cget(int, "-row")

    @row.setter
    def row(self, value: int) -> None:
        self._config(row=value)

    @property
    def col(self) -> int | None:
        return self._cget(int, "-column")

    @col.setter
    def col(self, value: int) -> None:
        self._config(column=value)

    @property
    def rowspan(self) -> int | None:
        return self._cget(int, "-rowspan")

    @rowspan.setter
    def rowspan(self, value: int) -> None:
        self._config(rowspan=value)

    @property
    def colspan(self) -> int | None:
        return self._cget(int, "-columnspan")

    @colspan.setter
    def colspan(self, value: int) -> None:
        self._config(columnspan=value)

    @property
    def cell(self) -> str | None:
        try:
            return self._widget.parent.grid._cell_managed_children[self._widget]
        except KeyError:
            return None

    @cell.setter
    def cell(self, value: str) -> None:
        self._set_cell(value)

    @property
    def margin(self) -> tuple[int, ...]:
        margin = [item if len(item) == 2 else item * 2 for item in self._get_pad()]
        return margin[1][0], margin[0][1], margin[1][1], margin[0][0]

    @margin.setter
    def margin(self, value: tuple[int, ...]) -> None:
        padx, pady = self._parse_margin(value)
        self._config(padx=padx, pady=pady)

    @property
    def align(self) -> tuple[Align | None, Align | None]:
        h = v = None
        sticky = set(self._cget(str, "-sticky"))

        if sticky & {"w", "e"} == {"w", "e"}:
            h = Align.Stretch
        elif "w" in sticky:
            h = Align.Start
        elif "e" in sticky:
            h = Align.End

        if sticky & {"n", "s"} == {"n", "s"}:
            v = Align.Stretch
        elif "n" in sticky:
            v = Align.Start
        elif "s" in sticky:
            v = Align.End

        return h, v

    @align.setter
    def align(self, value: tuple[Align | None, Align | None] | None) -> None:
        self._config(sticky=self._parse_align(value))


class Geometry(LayoutManager):
    # TODO: Do something with % placement
    # TODO: Properties

    _type = "place"

    def __call__(
        self,
        left: float | None = None,
        top: float | None = None,
        right: float | None = None,
        bottom: float | None = None,
    ):
        warnings.warn("This layout manager is WIP. Some things may not work.", Warning)

        x, y = left, top
        rely = relheight = height = None
        relx = relwidth = width = None

        if (x is not None and 0 < x < 1) or (y is not None and 0 < y < 1):
            x, relx = 0, x
            y, rely = 0, y
            anchor = "center"
        else:
            if bottom is not None:
                rely = 1
                y = -bottom

                if top is not None:
                    rely = 0
                    y = top
                    relheight = 1
                    height = -bottom - top

            if right is not None:
                relx = 1
                x = -right

                if left is not None:
                    relx = 0
                    x = left
                    relwidth = 1
                    width = -right - left

            anchor = self._get_anchor(x, y, relx, rely)

        Tcl.call(
            None,
            "place",
            "configure",
            self._widget._lm_path,
            *Tcl.to_tcl_args(
                x=x,
                y=y,
                relx=relx,
                rely=rely,
                width=width,
                height=height,
                relwidth=relwidth,
                relheight=relheight,
                anchor=anchor,
            ),
        )

    def _get_anchor(
        self, x: float | None, y: float | None, relx: float | None, rely: float | None
    ) -> str:
        anchor = ""

        if y is None:
            anchor += "n"
        elif y >= 0 and rely is not None and rely > 0 or y < 0:
            anchor += "s"
        else:
            anchor += "n"

        if x is None:
            anchor += "w"
        elif x >= 0 and relx is not None and relx > 0 or x < 0:
            anchor += "e"
        else:
            anchor += "w"

        return anchor


class Position(LayoutManager):
    _type = "place"

    def __call__(
        self,
        x: float | None = 0,
        y: float | None = 0,
        width: float | None = None,
        height: float | None = None,
        anchor: Anchor | None = None,
    ):
        Tcl.call(
            None,
            "place",
            "configure",
            self._widget._lm_path,
            *Tcl.to_tcl_args(
                x=x,
                y=y,
                width=width,
                height=height,
                anchor=anchor,
            ),
        )

    @property
    def x(self) -> int | None:
        return self._cget(int, "-x")

    @x.setter
    def x(self, value: int) -> None:
        self._config(x=value)

    @property
    def y(self) -> int | None:
        return self._cget(int, "-y")

    @y.setter
    def y(self, value: int) -> None:
        self._config(y=value)

    @property
    def width(self) -> int | None:
        return self._cget(int, "-width")

    @width.setter
    def width(self, value: int) -> None:
        self._config(width=value)

    @property
    def height(self) -> int | None:
        return self._cget(int, "-height")

    @height.setter
    def height(self, value: int) -> None:
        self._config(height=value)

    @property
    def anchor(self) -> int | None:
        return self._cget(Anchor, "-anchor")

    @anchor.setter
    def anchor(self, value: int) -> None:
        self._config(anchor=value)


class GridCells:
    @staticmethod
    def _parse(cells_list: list[list[str | None]]):
        result: dict[str, dict[str, int | bool]] = {}

        for row_index, row_list in enumerate(cells_list):
            for col_index, cell_name in enumerate(row_list):
                if cell_name is None:
                    continue
                if cell_name in result:
                    if not result[cell_name].get("cols_counted", False):
                        result[cell_name]["colspan"] = row_list.count(cell_name)
                        result[cell_name]["cols_counted"] = True

                    if not result[cell_name].get("rows_counted", False):
                        result[cell_name]["rowspan"] = sum(cell_name in item for item in cells_list)
                        result[cell_name]["rows_counted"] = True
                else:
                    result[cell_name] = {
                        "row": row_index,
                        "col": col_index,
                        "rowspan": 1,  # reset to 1 when modifying the cell layout
                        "colspan": 1,
                    }

        return result

    @staticmethod
    def _update(owner):
        for widget, cell in owner.grid._cell_managed_children.items():
            widget.grid._set_cell(cell)  # _set_cell does the update

    def __set__(self, obj, value: list[list[str | None]]) -> None:
        obj._cells = value
        obj._cells_values = GridCells._parse(value)
        GridCells._update(obj._widget)

    def __get__(self, obj, *_) -> list[list[str | None]]:
        return obj._cells


class ToplevelGrid:
    # TODO: row/col weight
    # TODO: row/col gap

    _cells = []
    _cells_values = {}
    _cell_managed_children = {}

    cells = GridCells()

    def __init__(self, owner: WindowBase) -> None:
        self._widget = owner

    @property
    def size(self) -> tuple[int, int]:
        return Tcl.call((int,), "grid", "size", self._widget)

    def detect(self, x: int, y: int):
        return tuple(reversed(Tcl.call((int,), "grid", "location", self._widget, x, y)))


class ContainerGrid(Grid, ToplevelGrid):
    ...



from __future__ import annotations

import re
from functools import partial
from pathlib import Path
from typing import Any, Callable

from PIL import Image, UnidentifiedImageError  # type: ignore

from ._info import System
from ._keysyms import _keysym_aliases
from ._tcl import Tcl
from ._utils import reversed_dict
from .colors import Color

if System.os == "macOS":
    button_numbers = {1: "left", 2: "right", 3: "middle", None: None}
    modifier_order = {"Alt_Opt": "⌥", "Shift": "⇧", "Ctrl_Ctrl": "⌃", "Ctrl_Cmd": "⌘"}
    platform_sortcuts = {"Preferences": ("Ctrl_Cmd", ","), "Help": ("Ctrl_Cmd", "?")}
    shortcut_separator = ""
else:
    button_numbers = {1: "left", 2: "middle", 3: "right", None: None}
    modifier_order = {"Ctrl_Ctrl": "Ctrl", "Ctrl_Cmd": "Ctrl", "Alt_Opt": "Alt", "Shift": "Shift"}
    platform_sortcuts = {"Preferences": ("Ctrl_Cmd", "."), "Help": ("F1",)}  # type: ignore  # mypy doesn't like ("F1",)
    shortcut_separator = "+"


class DataContainer:
    _container_dict: dict[str, Any] = {}

    def add(self, value: Any) -> str:
        key = str(id(value))
        self._container_dict[key] = value
        return key

    def pop(self, key: str) -> Any:
        return self._container_dict.pop(key, None)

    def __getitem__(self, key: str) -> Any:
        return self._container_dict[key]


_virtual_event_data_container = DataContainer()


class Event:
    binding_substitutions: dict[str, tuple[str, type | tuple[type]]] = {
        "button": ("%b", int),
        "modifiers": ("%s", int),
        "keycode": ("%k", int),
        "rel_x": (r"%x", int),
        "rel_y": (r"%y", int),
        "char": ("%A", str),
        "delta": ("%D", float),
        "keysymbol": ("%K", str),
        "abs_x": (r"%X", int),
        "abs_y": (r"%Y", int),
    }
    _modifiers = {
        0x1000: "MouseWheel:Down",
        0x800: "MouseWheel:Up",
        0x400: "MouseButton:Right",
        0x200: "MouseButton:Middle",
        0x100: "MouseButton:Left",
        0x80: "AltGr",
        0x40: "Super",
        0x20: "idk",  # TODO
        0x10: "NumLock",
        0x8: "Alt",
        0x4: "Control",
        0x2: "CapsLock",
        0x1: "Shift",
    }

    sequence: str
    sequences: dict[str, str]

    _ignored_values: tuple[Any, ...] = (None, "??", -1, 0)
    _order: tuple[str, ...] = tuple(binding_substitutions.keys())
    _relevant_attributes: tuple[str, ...] = ()
    _result = True

    @classmethod
    def _parse(cls, sequence: str) -> str:
        return cls.sequences[sequence]

    def __init__(self, *args) -> None:
        for item in self._relevant_attributes:
            value = args[self._order.index(item)]
            if value != "??":
                value = Tcl.from_(self.binding_substitutions[item][1], value)
            else:
                value = None
            setattr(self, item, value)

    def __repr__(self) -> str:
        pairs = []

        for name in self._relevant_attributes:
            value = getattr(self, name)
            if value not in self._ignored_values:
                pairs.append(f"{name}={value!r}")

        return f"<{type(self).__name__}: {', '.join(pairs)}>"

    def _get_modifiers(self, value: int) -> set[str]:
        result = []

        for modifier_value in self._modifiers:
            # tkinter uses some crazy bit shifting stuff here that makes me dizzy
            # imo my solution is much more clear
            if value >= modifier_value:
                value -= modifier_value
                result.append(self._modifiers[modifier_value])

        return set(result)

    def stop(self):
        self._result = False
        return self._result


class KeyboardEvent(Event):
    sequences = {
        "<KeyDown:Any>": "<KeyPress>",
        "<KeyDown>": "<KeyPress>",
        "<KeyUp:Any>": "<KeyRelease>",
        "<KeyUp>": "<KeyRelease>",
    }

    _ignored_values: tuple[Any, ...] = (None, "", "??", -1, 0, (), set())
    _regex_str = r"<Key(Down|Up):\((.*?)\)>"
    _relevant_attributes = ("char", "modifiers", "keysymbol", "keycode")
    _reversed_aliases = reversed_dict(_keysym_aliases)

    @classmethod
    def _matches(cls, sequence: str) -> bool:
        return sequence in cls.sequences or bool(re.match(cls._regex_str, sequence))

    @classmethod
    def _parse(cls, sequence: str) -> str:
        up_or_down = {"Down": "Press", "Up": "Release"}

        if sequence in cls.sequences:
            return cls.sequences[sequence]
        elif re.match(cls._regex_str, sequence):
            search = re.search(cls._regex_str, sequence)
            assert search is not None  # mypy

            keys = search[2].split("-")
            key = keys[-1]
            if key in cls._reversed_aliases:
                key = cls._reversed_aliases[key]

            modifier_keys = keys[:-1]
            for mod in modifier_keys:
                if mod in cls._reversed_aliases:
                    modifier_keys.remove(mod)
                    modifier_keys.append(cls._reversed_aliases[mod])

            if "Shift" in modifier_keys:
                if len(key) == 1:
                    key = key.upper()
                elif key == "Tab" and System.win_sys == "X11":
                    modifier_keys.remove("Shift")
                    key = "ISO_Left_Tab"

            modifiers = "-".join(modifier_keys) + ("-" if modifier_keys else "")

            return f"<{modifiers}Key{up_or_down[search[1]]}-{key}>"

        raise ValueError

    def __init__(self, *args) -> None:
        for item in self._relevant_attributes:
            value = args[self._order.index(item)]
            if value != "??":
                value = Tcl.from_(self.binding_substitutions[item][1], value)
            else:
                value = None

            if item == "keysymbol" and value in _keysym_aliases:
                value = _keysym_aliases[value]

            if item == "modifiers":
                value = self._get_modifiers(value)

            setattr(self, item, value)


class MouseEvent(Event):
    sequences = {
        # button press
        "<MouseDown:Any:Double>": "<Double-ButtonPress>",
        "<MouseDown:Any:Triple>": "<Triple-ButtonPress>",
        "<MouseDown:Any>": "<ButtonPress>",
        "<MouseDown:Left:Double>": "<Double-ButtonPress-1>",
        "<MouseDown:Left:Triple>": "<Triple-ButtonPress-1>",
        "<MouseDown:Left>": "<ButtonPress-1>",
        "<MouseDown:Middle:Double>": "<Double-ButtonPress-2>",
        "<MouseDown:Middle:Triple>": "<Triple-ButtonPress-2>",
        "<MouseDown:Middle>": "<ButtonPress-2>",
        "<MouseDown:Right:Double>": "<Double-ButtonPress-3>",
        "<MouseDown:Right:Triple>": "<Triple-ButtonPress-3>",
        "<MouseDown:Right>": "<ButtonPress-3>",
        "<MouseDown>": "<ButtonPress>",
        # drag
        "<MouseDrag:Left>": "<Button1-Motion>",
        "<MouseDrag:Middle>": "<Button2-Motion>",
        "<MouseDrag:Right>": "<Button3-Motion>",
        # button release
        "<MouseUp:Any>": "<ButtonRelease>",
        "<MouseUp:Left>": "<ButtonRelease-1>",
        "<MouseUp:Middle>": "<ButtonRelease-2>",
        "<MouseUp:Right>": "<ButtonRelease-3>",
        "<MouseUp>": "<ButtonRelease>",
        # mouse motion
        "<MouseEnter>": "<Enter>",
        "<MouseLeave>": "<Leave>",
        "<MouseMotion>": "<Motion>",
        # focus (i have no better idea where to put them)
        "<FocusOut>": "<FocusOut>",
        "<FocusIn>": "<FocusIn>",
    }
    _ignored_values: tuple[Any, ...] = (None, "??", -1, 0, set())
    _relevant_attributes = ("button", "modifiers", "abs_x", "abs_y", "rel_x", "rel_y")

    def __init__(self, *args) -> None:
        for item in self._relevant_attributes:
            value = args[self._order.index(item)]
            if value != "??":
                value = Tcl.from_(self.binding_substitutions[item][1], value)
            else:
                value = None

            if value and item == "button":
                value = button_numbers[value]

            if item == "modifiers":
                value = self._get_modifiers(value)

            setattr(self, item, value)


class ScrollEvent(Event):
    _ignored_values: tuple[Any, ...] = (None, "??", set())
    _relevant_attributes = ("delta", "modifiers")

    @classmethod
    def _matches(cls, sequence) -> bool:
        return any(i in sequence for i in ("MouseWheel", "Button-4", "Button-5"))

    @classmethod
    def _parse(cls, sequence: str) -> str:
        return sequence

    def __init__(self, *args) -> None:
        if System.win_sys == "X11":
            num = Tcl.from_(int, args[self._order.index("button")])
            self.delta = -1 if num == 4 else 1
        else:
            self.delta = Tcl.from_(int, args[self._order.index("delta")])

        self.modifiers = self._get_modifiers(Tcl.from_(int, args[self._order.index("modifiers")]))


class VirtualEvent(Event):
    binding_substitutions = {"data": ("%d", int)}

    @classmethod
    def _matches(cls, sequence) -> bool:
        return sequence.startswith("<<") and sequence.endswith(">>") and len(sequence) > 4

    @classmethod
    def _parse(cls, sequence: str) -> str:
        return sequence

    def __init__(self, *args) -> None:
        data_key = Tcl.from_(str, args[0])

        self.data = _virtual_event_data_container.pop(data_key) if data_key else None

    def __repr__(self) -> str:
        return f"<VirtualEvent: {self.sequence.strip('<').strip('>')}; data={self.data!r}>"


class DnDEvent(Event):
    binding_substitutions = {
        "action": ("%A", str),
        "button": ("%b", int),
        "common_formats": ("%CST", (str,)),
        "_data": ("%D", str),
        "format": ("%T", str),
        "formats": ("%L", (str,)),
        "type": ("%CPT", str),
        "rel_x": (r"%X", int),
        "rel_y": (r"%Y", int),
    }
    sequences = {
        # drop
        "<<Drop:Any>>": "<<Drop:*>>",
        "<<Drop:App>>": "<<Drop:DND_App>>",
        "<<Drop:Color>>": "<<Drop:DND_Color>>",
        "<<Drop:Custom>>": "<<Drop:DND_UserDefined>>",
        "<<Drop:File>>": "<<Drop:DND_Files>>",
        "<<Drop:HTML>>": "<<Drop:DND_HTML>>",
        "<<Drop:RTF>>": "<<Drop:DND_RTF>>",
        "<<Drop:Text>>": "<<Drop:DND_Text>>",
        "<<Drop:URL>>": "<<Drop:DND_URL>>",
        "<<Drop>>": "<<Drop>>",
        # drag
        "<<DragEnd>>": "<<DragEndCmd>>",
        "<<DragEnter>>": "<<DropEnter>>",
        "<<DragLeave>>": "<<DropLeave>>",
        "<<DragMotion>>": "<<DropPosition>>",
        "<<DragStart>>": "<<DragInitCmd>>",
    }

    _existing_types = {
        # cross-platform
        "DND_App",
        "DND_Color",
        "DND_Files",
        "DND_HTML",
        "DND_RTF",
        "DND_Text",
        "DND_URL",
        # tkdnd_macosx.tcl
        "NSPasteboardTypeString",
        "NSFilenamesPboardType",
        "NSPasteboardTypeHTML",
        # tkdnd_windows.tcl
        "CF_UNICODETEXT",
        "CF_TEXT",
        "CF_HDROP",
        "UniformResourceLocator",
        "CF_HTML",
        "HTML Format",
        "CF_RTF",
        "CF_RTFTEXT",
        "Rich Text Format",
        # tkdnd_unix.tcl
        r"text/plain\;charset=utf-8",
        "UTF8_STRING",
        "text/plain",
        "text/rtf",
        "text/richtext",
        "STRING",
        "TEXT",
        "COMPOUND_TEXT",
        "text/uri-list",
        r"text/html\;charset=utf-8",
        "text/html",
        "application/x-color",
        # add more of these app-specific types?
        # tkdnd raises an error, if there's no common types between source and dest
        "application/x-orgkdeplasmataskmanager_taskbuttonitem",
        "text/x-orgkdeplasmataskmanager_taskurl",
    }
    _dnd_type_aliases = {
        "app": "DND_App",
        "color": "DND_Color",
        "custom": "DND_UserDefined",
        "file": "DND_Files",
        "html": "DND_HTML",
        "rtf": "DND_RTF",
        "text": "DND_Text",
        "url": "DND_URL",
    }
    _result = "copy"

    type: str
    format: str
    action: str

    def __init__(self, *args) -> None:
        result = {}

        for (attr, (_, type_)), value in zip(self.binding_substitutions.items(), args):
            value = Tcl.from_(type_, value) if value != "??" else None
            result[attr] = value

        try:
            dropped_data_type = result["type"] = reversed_dict(self._dnd_type_aliases)[
                result["type"]
            ]
        except KeyError:
            dropped_data_type = None

        self._needs_download = False
        self._image_regex = re.compile(r'(.*?)<img (.*?)src="(.*?)"(.*?)>(.*?)', re.DOTALL)

        if dropped_data_type == "file":
            data = Tcl.from_([str], result["_data"])

            result["_data"] = [Path(x) for x in data] if len(data) > 1 else Path(data[0])
        elif dropped_data_type == "html":
            # search for an image url in the dropped html, and return a PIL.Image.Image object
            if re.match(self._image_regex, result["_data"]):
                result["_data"]
                self._needs_download = True

        elif dropped_data_type == "color":
            color = "#"
            splitted_color = Tcl.from_([str], result["_data"])

            if len(splitted_color) == 4:
                for i in splitted_color[:3]:
                    color += i[4:]
            else:
                color = result["_data"]

            result["_data"] = Color(color)

        elif dropped_data_type == "custom":
            ord_list = Tcl.from_([str], result["_data"])
            char_tuple = tuple(
                chr(int(x)) for x in ord_list
            )  # faster to int() here than in Tcl.from_

            result["_data"] = "".join(char_tuple)

        result["button"] = button_numbers[result["button"]]

        for key, value in result.items():
            setattr(self, key, value)

    def __repr__(self) -> str:
        if self.type == "custom":
            more_useful_type_info = f", format={self.format!r}"
        else:
            more_useful_type_info = f", type={self.type!r}"

        if "Drop" in self.sequence:
            data_or_xy = f"data={self.data!r}"
        else:
            data_or_xy = f"rel_x={self.rel_x}, rel_y={self.rel_y}"

        return f"<DnDEvent: {data_or_xy}, action={self.action!r}{more_useful_type_info}>"

    @property
    def data(self) -> str | list | Image.Image | Path | Color:
        if self._needs_download:
            import tempfile

            import requests

            try:
                path = re.search(self._image_regex, self._data)[3]  # type: ignore
                _, name = tempfile.mkstemp()

                with open(name, "wb") as file:
                    file.write(requests.get(path).content)

                self._data = Image.open(name)
            except requests.exceptions.ConnectionError:
                pass
            except UnidentifiedImageError:
                self._needs_download = False
            else:
                self._needs_download = False

        return self._data

    def ignore(self):
        self._result = "refuse_drop"
        return self._result

    def move(self):
        self._result = "move"
        return self._result

    def copy(self):
        self._result = "copy"
        return self._result

    def link(self):
        self._result = "link"
        return self._result

    def stop(self):
        self._result = "refuse_drop"
        return self._result


class EventMixin:
    _name: str

    def _bind(self, sequence: str, func: Callable | str, overwrite: bool, send_event: bool) -> None:
        event: type[KeyboardEvent | MouseEvent | ScrollEvent | DnDEvent | VirtualEvent]

        if KeyboardEvent._matches(sequence):
            event = KeyboardEvent
        elif sequence in MouseEvent.sequences:
            event = MouseEvent
        elif ScrollEvent._matches(sequence):
            event = ScrollEvent
        elif sequence in DnDEvent.sequences:
            event = DnDEvent
        elif VirtualEvent._matches(sequence):
            event = VirtualEvent
        else:
            raise ValueError(f"invalid event sequence: {sequence}")
        _orig_seq = sequence

        def real_func(func, *args):
            if not send_event:
                return func()

            event_to_send = event(*args)
            event_to_send.sequence = _orig_seq
            result = func(event_to_send)

            if result is not None:
                return result

            return event_to_send._result

        script_str = func  # for internal bindings, where i use strings for tcl scripts, or when unbinding where func is ""

        if callable(func):
            cmd = Tcl.create_cmd(partial(real_func, func))

            sequence = event._parse(sequence)

            subst_str = " ".join(item[0] for item in event.binding_substitutions.values())

            if event is DnDEvent:
                script_str = (
                    f"{'' if overwrite else '+'} {cmd} {subst_str}"  # tcl: {+ command %subst}
                )
            else:
                script_str = f"{'' if overwrite else '+'} if {{[{cmd} {subst_str}] == 0}} break"  # tcl: {+ if {[command %subst] == 0} break}

        Tcl.call(None, "bind", self._name if self._name != ".app" else ".", sequence, script_str)

    def bind(
        self,
        sequence: str,
        func: str | Callable[[], bool | None] | Callable[[Event], bool | None],
        *,
        overwrite: bool = False,
        send_event: bool = False,
    ) -> None:
        if isinstance(sequence, KeySeq):
            sequence = sequence.event_sequence

        if "MouseScroll" in sequence:
            if System.win_sys == "X11":
                self._bind(sequence.replace("MouseScroll", "Button-4"), func, overwrite, send_event)
                self._bind(sequence.replace("MouseScroll", "Button-5"), func, overwrite, send_event)
                return
            else:
                sequence = sequence.replace("MouseScroll", "MouseWheel")

        if "Menu" in sequence and System.os == "Windows":
            sequence = sequence.replace("Menu", "App")

        self._bind(sequence, func, overwrite, send_event)

    def unbind(self, sequence: str) -> None:
        self._bind(sequence, "", True, False)

    def generate_event(self, sequence: str, data: Any = None) -> None:
        global _virtual_event_data_container
        key = _virtual_event_data_container.add(data) if data is not None else None

        Tcl.call(None, "event", "generate", self, sequence, *Tcl.to_tcl_args(data=key))


def DragObject(
    data: str | list | tuple | Path | Color, type_: str | None = None, action: str = "copy"
) -> tuple[str, str, str]:
    if type_ is None:
        if isinstance(data, str):
            type_ = "DND_Text"
        elif isinstance(data, (Path, tuple, list)):  # List or tuple of Paths
            type_ = "DND_Files"
        elif isinstance(data, Color):
            type_ = "DND_Color"
        else:
            raise ValueError("can't detect DnD type automatically. Please specify it!")
    else:
        try:
            type_ = DnDEvent._dnd_type_aliases[type_]
        except KeyError:
            if type_ not in DnDEvent._existing_types:
                Tcl.eval(
                    None,
                    f"lappend tkdnd::generic::_platform2tkdnd {type_} DND_UserDefined"
                    + "\n"
                    + f"lappend tkdnd::generic::_tkdnd2platform [dict get $tkdnd::generic::_platform2tkdnd {type_}] {type_}",
                )

    return (action, type_, Tcl.to(data))


class KeySeq:
    _wrong_modifiers = {
        "Alt": "'Alt_Opt'",
        "Command": "'Ctrl_Cmd'",
        "Control": "'Ctrl_Cmd' or 'Ctrl_Ctrl'",
    }
    _preset_shortcuts = {
        "bold": ("Ctrl_Cmd", "b"),
        "close": ("Ctrl_Cmd", "w"),
        "copy": ("Ctrl_Cmd", "c"),
        "cut": ("Ctrl_Cmd", "x"),
        "find": ("Ctrl_Cmd", "f"),
        "italic": ("Ctrl_Cmd", "i"),
        "new": ("Ctrl_Cmd", "n"),
        "open": ("Ctrl_Cmd", "o"),
        "paste": ("Ctrl_Cmd", "v"),
        "print": ("Ctrl_Cmd", "p"),
        "quit": ("Ctrl_Cmd", "q"),
        "redo": ("Ctrl_Cmd", "y"),
        "refresh": ("F5",),
        "save as": ("Ctrl_Cmd", "Shift", "s"),
        "save": ("Ctrl_Cmd", "s"),
        "underline": ("Ctrl_Cmd", "u"),
        "undo": ("Ctrl_Cmd", "z"),
        "zoom in": ("Ctrl_Cmd", "+"),
        "zoom out": ("Ctrl_Cmd", "-"),
    }
    _preset_shortcuts.update(platform_sortcuts)

    def __init__(self, *args) -> None:
        if args[0] in self._preset_shortcuts:
            self._sequence = list(self._preset_shortcuts[args[0]])
        else:
            seq = []
            real_key = ""

            for key in list(args):
                if key in self._wrong_modifiers:
                    raise ValueError(
                        f"wrong modifier key: {key!r}. Use {self._wrong_modifiers[key]} instead."
                    )
                elif " " in key:
                    raise ValueError(f"invalid key: {key!r}")

                if key in modifier_order:
                    seq.append(key)
                else:
                    real_key = key

            seq.append(real_key)
            self._sequence = seq

    @property
    def event_sequence(self) -> str:
        return f"<KeyDown:({'-'.join(self._sequence)})>"

    @property
    def shortcut_sequence(self) -> str:
        real_key = self._sequence[-1]

        result = [symbol for name, symbol in modifier_order.items() if name in self._sequence]

        result.append(real_key.upper())

        return shortcut_separator.join(result)

    def __eq__(self, other: object) -> bool:
        return (
            set(self._sequence) == set(other._sequence)
            if isinstance(other, KeySeq)
            else NotImplemented
        )



from __future__ import annotations

import contextlib
from pathlib import Path

from PIL import Image as PIL_Image  # type: ignore

from ._base import WidgetBase
from ._props import cget, config
from ._tcl import Tcl
from ._utils import _images, _pil_images, counts
from .colors import Color

try:
    from PIL import Image as PillowImage
    from PIL import _imagingtk as ImagingTk  # type: ignore
except ImportError as e:
    raise ImportError("Tukaan needs PIL and PIL._imagingtk in order to display images.") from e

PHOTO_CMD = "image create photo {name}\nPyImagingPhoto {name} {blockid}"


class Pillow2Tcl:
    _cur_f = 0

    def __init__(self, image: PillowImage.Image):
        if image in _pil_images.values():
            for key, value in _pil_images.items():
                if value is image:
                    self._name = key
                    return

        try:
            _animated = image.is_animated
        except AttributeError:
            _animated = False

        self._transparent = "transparency" in image.info
        self._image = image
        self._name = f"tukaan_image_{next(counts['images'])}"

        Tcl.call(None, "image", "create", "photo", self._name)
        ImagingTk.tkinit(Tcl.interp_address, 1)

        _images[self._name] = self  # gc
        _pil_images[self._name] = image

        if _animated:
            self.frames = []
            self.show_cmd = Tcl.create_cmd(self._show)
            self._start()
        else:
            self._create(self._name, self._image)

    def _create(self, name: str, image: PillowImage.Image) -> tuple[str, int]:
        if not hasattr(image, "mode") and not hasattr(image, "size"):
            raise

        if self._transparent:
            image = image.convert("RGBA")

        mode = image.mode
        image.load()

        if mode == "P":
            try:
                mode = image.palette.mode
                if mode not in {"1", "L", "RGB", "RGBA"}:
                    mode = PillowImage.getmodebase(mode)
            except AttributeError:
                mode = "RGB"

        try:
            duration = int(image.info["duration"])
        except KeyError:
            duration = 50

        size = image.size
        im = image.im

        if im.isblock():
            block = im
        else:
            block = im.new_block(mode, size)
            im.convert2(block, im)

        Tcl.eval(None, PHOTO_CMD.format(name=name, blockid=block.id))

        return name, duration

    def _start(self) -> None:
        frame_count = 0
        with contextlib.suppress(EOFError):
            while True:
                name = f"{self._name}_frame_{frame_count}"
                self._image.seek(frame_count)
                self.frames.append(self._create(name, self._image))
                frame_count += 1

        self._len = frame_count - 1
        self._show()

    def _show(self) -> None:
        name, duration = self.frames[self._cur_f]

        if self._cur_f != self._len:
            self._cur_f += 1
        else:
            self._cur_f = 0

        Tcl.eval(
            None,
            f"{self._name} copy {name} -compositingrule set\n"
            + f"after {duration} {self.show_cmd}",
        )

    def __to_tcl__(self) -> str:
        return self._name

    @classmethod
    def __from_tcl__(cls, value: str) -> PIL_Image.Image | None:
        if isinstance(value, (tuple, list)):
            [value] = value
        return _pil_images.get(value)


def pil_image_to_tcl(self):
    return Pillow2Tcl(self).__to_tcl__()


setattr(PillowImage.Image, "__to_tcl__", pil_image_to_tcl)


class Icon:
    def __init__(self, file: str | Path | None = None) -> None:
        self._name = f"tukaan_icon_{next(counts['icons'])}"
        _images[self._name] = self

        Tcl.call(
            None,
            "image",
            "create",
            "photo",
            self._name,
            *Tcl.to_tcl_args(file=file),
        )

    def config(self, file: Path | None = None):
        Tcl.call(
            None,
            self._name,
            "configure",
            *Tcl.to_tcl_args(file=file),
        )

    def __to_tcl__(self) -> str:
        return self._name

    @classmethod
    def __from_tcl__(cls, value: str) -> Icon:
        return _images[value]  # type: ignore


class IconFactory:
    def __init__(self, on_light: Path, on_dark: Path | None):
        self._light_icons_dir = on_light
        self._dark_icons_dir = on_dark
        self._current_dir = on_light
        self.cache: dict[str, Icon] = {}

        self._current_dir = on_light

        if on_dark is not None:
            Tcl.call(None, "bind", ".app", "<<ThemeChanged>>", self.change_icon_theme)

    def change_icon_theme(self):
        is_dark = Color(
            Tcl.call(str, "ttk::style", "lookup", "TLabel.label", "-foreground")
        ).is_dark

        self._current_dir = self._light_icons_dir if is_dark else self._dark_icons_dir
        assert self._current_dir is not None

        for name, icon in self.cache.items():
            icon.config(file=Path(self._current_dir) / f"{name}.png")

    def get(self, icon_name: str) -> Icon:
        if icon_name in self.cache:
            return self.cache[icon_name]

        assert self._current_dir is not None
        icon = Icon(file=Path(self._current_dir) / f"{icon_name}.png")
        self.cache[icon_name] = icon
        return icon

    __getitem__ = get


def get_image(self) -> PillowImage.Image | Icon:
    return cget(self, Pillow2Tcl, "-image")


def set_image(self, value: PillowImage.Image | Icon) -> None:
    config(self, image=value)


image = property(get_image, set_image)


class Image(WidgetBase):
    _tcl_class = "ttk::label"

    image = image

    def __init__(self, parent, image: PillowImage.Image | None = None):
        WidgetBase.__init__(self, parent, image=image)



from __future__ import annotations

from typing import Callable

from tukaan._base import InputControl, TkWidget, WidgetBase
from tukaan._props import config, focusable, orientation
from tukaan._tcl import Tcl
from tukaan.enums import Orientation
from tukaan.exceptions import WidgetError


class ScrollBar(WidgetBase, InputControl):
    _tcl_class = "ttk::scrollbar"

    focusable = focusable
    orientation = orientation

    _hidden = False

    def __init__(
        self,
        parent: TkWidget,
        orientation: Orientation | None = None,
        *,
        auto_hide: bool = True,
        focusable: bool | None = None,
        tooltip: str | None = None,
    ) -> None:
        self._auto_hide = auto_hide

        WidgetBase.__init__(self, parent, orient=orientation, takefocus=focusable, tooltip=tooltip)

    def _set_command(self, value: Callable) -> None:
        config(self, command=value)

    def _set_range(self, start: str, end: str) -> None:
        if self._auto_hide:
            should_be_hidden = float(start) <= 0.0 and float(end) >= 1.0

            if should_be_hidden and not self._hidden:
                self.hide()
                self._hidden = True
            elif not should_be_hidden and self._hidden:
                self.unhide()
                self._hidden = False

        Tcl.call(None, self, "set", start, end)

    def attach(self, widget: WidgetBase) -> None:
        if self.orientation is Orientation.Horizontal:
            try:
                widget.on_xscroll = self._set_range
                self._set_command(widget.x_scroll)
            except AttributeError as e:
                raise WidgetError("this widget isn't scrollable horizontally") from e

        elif self.orientation is Orientation.Vertical:
            try:
                widget.on_yscroll = self._set_range
                self._set_command(widget.y_scroll)
            except AttributeError as e:
                raise WidgetError("this widget isn't scrollable vertically") from e

    @property
    def auto_hide(self) -> bool:
        return self._auto_hide

    @auto_hide.setter
    def auto_hide(self, value: bool) -> None:
        self._auto_hide = value



_keysym_aliases = {
    # general
    "Alt_L": "Alt:Left",
    "Alt_R": "Alt:Right",
    "App": "Menu",
    "BackSpace": "Backspace",
    "Caps_Lock": "CapsLock",
    "Control_L": "Control:Left",
    "Control_R": "Control:Right",
    "Hyper_L": "Hyper:Left",
    "Hyper_R": "Hyper:Right",
    "ISO_Left_Tab": "Tab",
    "ISO_Level3_Shift": "AltGr",
    "Meta_L": "Meta:Left",
    "Meta_R": "Meta:Right",
    "Mode_switch": "ModeSwitch",
    "Next": "PageDown",
    "Num_Lock": "NumLock",
    "Pause": "PauseBreak",
    "Print": "PrintScreen",
    "Prior": "PageUp",
    "Return": "Enter",
    "Scroll_Lock": "ScrollLock",
    "Shift_L": "Shift:Left",
    "Shift_Lock": "ShiftLock",
    "Shift_R": "Shift:Right",
    "Super_L": "Super:Left",
    "Super_R": "Super:Right",
    "Win_L": "Windows:Left",
    "Win_R": "Windows:Right",
    "Kana_Lock": "KanaLock",
    "Kana_Shift": "KanaShift",
    "space": "Space",
    # numpad
    "KP_0": "NumPad:0",
    "KP_1": "NumPad:1",
    "KP_2": "NumPad:2",
    "KP_3": "NumPad:3",
    "KP_4": "NumPad:4",
    "KP_5": "NumPad:5",
    "KP_6": "NumPad:6",
    "KP_7": "NumPad:7",
    "KP_8": "NumPad:8",
    "KP_9": "NumPad:9",
    "KP_Add": "NumPad:Add",
    "KP_Begin": "NumPad:Begin",
    "KP_Decimal": "NumPad:Decimal",
    "KP_Delete": "NumPad:Delete",
    "KP_Divide": "NumPad:Divide",
    "KP_Down": "NumPad:Down",
    "KP_End": "NumPad:End",
    "KP_Enter": "NumPad:Enter",
    "KP_Equal": "NumPad:Equal",
    "KP_F1": "NumPad:F1",
    "KP_F2": "NumPad:F2",
    "KP_F3": "NumPad:F3",
    "KP_F4": "NumPad:F4",
    "KP_Home": "NumPad:Home",
    "KP_Insert": "NumPad:Insert",
    "KP_Left": "NumPad:Left",
    "KP_Multiply": "NumPad:Multiply",
    "KP_Next": "NumPad:PageDown",
    "KP_Prior": "NumPad:PageUp",
    "KP_Right": "NumPad:Right",
    "KP_Separator": "NumPad:Separator",
    "KP_Space": "NumPad:Space",
    "KP_Subtract": "NumPad:Substract",
    "KP_Tab": "NumPad:Tab",
    "KP_Up": "NumPad:Up",
    # XF86 stuff
    "XF86AddFavorite": "XF86AddFavorite",
    "XF86ApplicationLeft": "XF86ApplicationLeft",
    "XF86ApplicationRight": "XF86ApplicationRight",
    "XF86AudioCycleTrack": "XF86AudioCycleTrack",
    "XF86AudioForward": "Audio:Forward",
    "XF86AudioLowerVolume": "Audio:VolumeDown",
    "XF86AudioMedia": "XF86AudioMedia",
    "XF86AudioMicMute": "XF86AudioMicMute",
    "XF86AudioMute": "Audio:Mute",
    "XF86AudioNext": "Audio:Next",
    "XF86AudioPause": "Audio:Pause",
    "XF86AudioPlay": "Audio:Play",
    "XF86AudioPreset": "XF86AudioPreset",
    "XF86AudioPrev": "Audio:Prev",
    "XF86AudioRaiseVolume": "Audio:VolumeUp",
    "XF86AudioRandomPlay": "Audio:Shuffle",
    "XF86AudioRecord": "Audio:Record",
    "XF86AudioRepeat": "Audio:Repeat",
    "XF86AudioRewind": "Audio:Rewind",
    "XF86AudioStop": "Audio:Stop",
    "XF86Away": "XF86Away",
    "XF86Back": "Action:Back",
    "XF86BackForward": "XF86BackForward",
    "XF86Battery": "XF86Battery",
    "XF86Blue": "Color:Blue",
    "XF86Bluetooth": "XF86Bluetooth",
    "XF86Book": "XF86Book",
    "XF86BrightnessAdjust": "XF86BrightnessAdjust",
    "XF86CD": "XF86CD",
    "XF86Calculator": "App:Calculator",
    "XF86Calendar": "App:Calendar",
    "XF86Clear": "Action:Clear",
    "XF86ClearGrab": "XF86ClearGrab",
    "XF86Close": "Action:Close",
    "XF86Community": "XF86Community",
    "XF86ContrastAdjust": "XF86ContrastAdjust",
    "XF86Copy": "Action:Copy",
    "XF86Cut": "Action:Cut",
    "XF86CycleAngle": "XF86CycleAngle",
    "XF86DOS": "XF86DOS",
    "XF86Display": "XF86Display",
    "XF86Documents": "Place:Documents",
    "XF86Eject": "Eject",
    "XF86Excel": "App:Excel",
    "XF86Explorer": "App:Explorer",
    "XF86Favorites": "Favorites",
    "XF86Finance": "XF86Finance",
    "XF86Forward": "Action:Forward",
    "XF86FrameBack": "XF86FrameBack",
    "XF86FrameForward": "XF86FrameForward",
    "XF86Game": "Game",
    "XF86Go": "Go",
    "XF86Green": "Color:Green",
    "XF86Hibernate": "Power:Hibernate",
    "XF86History": "History",
    "XF86HomePage": "HomePage",
    "XF86HotLinks": "XF86HotLinks",
    "XF86KbdBrightnessDown": "Keyboard:BrightnessDown",
    "XF86KbdBrightnessUp": "Keyboard:BrightnessUp",
    "XF86KbdLightOnOff": "Keyboard:BacklightToggle",
    "XF86Keyboard": "Keyboard",
    "XF86Launch0": "XF86Launch0",
    "XF86Launch1": "XF86Launch1",
    "XF86Launch2": "XF86Launch2",
    "XF86Launch3": "XF86Launch3",
    "XF86Launch4": "XF86Launch4",
    "XF86Launch5": "XF86Launch5",
    "XF86Launch6": "XF86Launch6",
    "XF86Launch7": "XF86Launch7",
    "XF86Launch8": "XF86Launch8",
    "XF86Launch9": "XF86Launch9",
    "XF86LaunchA": "XF86LaunchA",
    "XF86LaunchB": "XF86LaunchB",
    "XF86LaunchC": "XF86LaunchC",
    "XF86LaunchD": "XF86LaunchD",
    "XF86LaunchE": "XF86LaunchE",
    "XF86LaunchF": "XF86LaunchF",
    "XF86LightBulb": "XF86LightBulb",
    "XF86LogGrabInfo": "XF86LogGrabInfo",
    "XF86LogOff": "Power:Logoff",
    "XF86LogWindowTree": "XF86LogWindowTree",
    "XF86Mail": "App:Mail",
    "XF86MailForward": "XF86MailForward",
    "XF86Market": "XF86Market",
    "XF86Meeting": "Meeting",
    "XF86Memo": "XF86Memo",
    "XF86MenuKB": "XF86MenuKB",
    "XF86MenuPB": "XF86MenuPB",
    "XF86Messenger": "App:Messages",
    "XF86ModeLock": "XF86ModeLock",
    "XF86MonBrightnessCycle": "Monitor:BrightnessCycle",
    "XF86MonBrightnessDown": "Monitor:BrightnessDown",
    "XF86MonBrightnessUp": "Monitor:BrightnessUp",
    "XF86Music": "Place:Music",
    "XF86MyComputer": "Place:ThisPC",
    "XF86MySites": "XF86MySites",
    "XF86New": "Action:New",
    "XF86News": "App:News",
    "XF86Next_VMode": "XF86Next_VMode",
    "XF86OfficeHome": "XF86OfficeHome",
    "XF86Open": "Action:Open",
    "XF86OpenURL": "XF86OpenURL",
    "XF86Option": "XF86Option",
    "XF86Paste": "Action:Paste",
    "XF86Phone": "Phone",
    "XF86Pictures": "Place:Pictures",
    "XF86PowerDown": "Power:ShutDown",
    "XF86PowerOff": "Power:PowerOff",
    "XF86Prev_VMode": "XF86Prev_VMode",
    "XF86Q": "XF86Q",
    "XF86RFKill": "XF86RFKill",
    "XF86Red": "Color:Red",
    "XF86Refresh": "Action:Refresh",
    "XF86Reload": "Action:Reload",
    "XF86Reply": "Action:Reply",
    "XF86RockerDown": "XF86RockerDown",
    "XF86RockerEnter": "XF86RockerEnter",
    "XF86RockerUp": "XF86RockerUp",
    "XF86RotateWindows": "XF86RotateWindows",
    "XF86RotationKB": "XF86RotationKB",
    "XF86RotationLockToggle": "XF86RotationLockToggle",
    "XF86RotationPB": "XF86RotationPB",
    "XF86Save": "Action:Save",
    "XF86ScreenSaver": "Power:ScreenSaver",
    "XF86ScrollClick": "Scroll:Click",
    "XF86ScrollDown": "Scroll:Down",
    "XF86ScrollUp": "Scroll:Up",
    "XF86Search": "Action:Search",
    "XF86Select": "Action:Select",
    "XF86Send": "Action:Send",
    "XF86Shop": "XF86Shop",
    "XF86Sleep": "Power:Sleep",
    "XF86Spell": "XF86Spell",
    "XF86SplitScreen": "Action:SplitScreen",
    "XF86Standby": "Power:Standby",
    "XF86Start": "Action:Start",
    "XF86Stop": "Action:Stop",
    "XF86Subtitle": "XF86Subtitle",
    "XF86Support": "XF86Support",
    "XF86Suspend": "Power:Suspend",
    "XF86Switch_VT_1": "XF86Switch_VT_1",
    "XF86Switch_VT_10": "XF86Switch_VT_10",
    "XF86Switch_VT_11": "XF86Switch_VT_11",
    "XF86Switch_VT_12": "XF86Switch_VT_12",
    "XF86Switch_VT_2": "XF86Switch_VT_2",
    "XF86Switch_VT_3": "XF86Switch_VT_3",
    "XF86Switch_VT_4": "XF86Switch_VT_4",
    "XF86Switch_VT_5": "XF86Switch_VT_5",
    "XF86Switch_VT_6": "XF86Switch_VT_6",
    "XF86Switch_VT_7": "XF86Switch_VT_7",
    "XF86Switch_VT_8": "XF86Switch_VT_8",
    "XF86Switch_VT_9": "XF86Switch_VT_9",
    "XF86TaskPane": "XF86TaskPane",
    "XF86Terminal": "App:Terminal",
    "XF86Time": "XF86Time",
    "XF86ToDoList": "XF86ToDoList",
    "XF86Tools": "Tools",
    "XF86TopMenu": "XF86TopMenu",
    "XF86TouchpadOff": "Touchpad:Off",
    "XF86TouchpadOn": "Touchpad:On",
    "XF86TouchpadToggle": "Touchpad:Toggle",
    "XF86Travel": "XF86Travel",
    "XF86UWB": "XF86UWB",
    "XF86Ungrab": "XF86Ungrab",
    "XF86User1KB": "XF86User1KB",
    "XF86User2KB": "XF86User2KB",
    "XF86UserPB": "XF86UserPB",
    "XF86VendorHome": "XF86VendorHome",
    "XF86Video": "XF86Video",
    "XF86View": "XF86View",
    "XF86WLAN": "XF86WLAN",
    "XF86WWAN": "XF86WWAN",
    "XF86WWW": "XF86WWW",
    "XF86WakeUp": "XF86WakeUp",
    "XF86WebCam": "XF86WebCam",
    "XF86WheelButton": "XF86WheelButton",
    "XF86Word": "App:Word",
    "XF86Xfer": "XF86Xfer",
    "XF86Yellow": "Color:Yellow",
    "XF86ZoomIn": "Action:ZoomIn",
    "XF86ZoomOut": "Action:ZoomOut",
    "XF86iTouch": "iTouch",
}



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from fractions import Fraction

from .exceptions import ColorError


class ColorSpace(ABC):
    @staticmethod
    @abstractmethod
    def convert_from():
        pass

    @staticmethod
    @abstractmethod
    def convert_to(*args):
        pass


class Hex(ColorSpace):
    @staticmethod
    def convert_from(string):
        int_value = int(string.lstrip("#"), 16)
        return int_value >> 16, int_value >> 8 & 0xFF, int_value & 0xFF

    @staticmethod
    def convert_to(r, g, b):
        return f"#{r:02x}{g:02x}{b:02x}"


class Rgb(ColorSpace):
    @staticmethod
    def convert_from(r, g, b):
        return r, g, b

    @staticmethod
    def convert_to(r, g, b):
        return r, g, b


class Hsv(ColorSpace):
    @staticmethod
    def convert_from(h, s, v):
        h, s, v = h / 360, s / 100, v / 100

        if s == 0.0:
            return tuple(round(x * 255) for x in (v, v, v))

        i = int(h * 6.0)
        f = (h * 6.0) - i

        p = v * (1.0 - s)
        q = v * (1.0 - s * f)
        t = v * (1.0 - s * (1.0 - f))

        r, g, b = (
            (v, t, p),
            (q, v, p),
            (p, v, t),
            (p, q, v),
            (t, p, v),
            (v, p, q),
        )[int(i % 6)]

        return tuple(round(x * 255) for x in (r, g, b))

    @staticmethod
    def convert_to(r, g, b):
        r, g, b = tuple(x / 255 for x in (r, g, b))

        high = max(r, g, b)
        low = min(r, g, b)
        diff = high - low

        h = 0
        s = 0 if high == 0 else (diff / high) * 100
        v = high * 100

        if high == r:
            h = (60 * ((g - b) / diff) + 360) % 360
        elif high == g:
            h = (60 * ((b - r) / diff) + 120) % 360
        elif high == b:
            h = (60 * ((r - g) / diff) + 240) % 360

        return tuple(round(x) for x in (h, s, v))


class Hsl(ColorSpace):
    @staticmethod
    def convert_from(h, s, l):
        h, s, l = h / 360, s / 100, l / 100

        if s == 0.0:
            return (round(l * 255),) * 3

        if l >= 0.5:
            tmp_1 = l + s - l * s
        elif l < 0.5:
            tmp_1 = l * (1 + s)

        tmp_2 = 2 * l - tmp_1

        def _(h):
            h = h % 1
            if h < 1 / 6:
                return tmp_2 + (tmp_1 - tmp_2) * h * 6
            if h < 0.5:
                return tmp_1
            if h < 2 / 3:
                return tmp_2 + (tmp_1 - tmp_2) * (2 / 3 - h) * 6
            return tmp_2

        r = _(h + 1 / 3)
        g = _(h)
        b = _(h - 1 / 3)

        return tuple(round(x) for x in (r * 255, g * 255, b * 255))

    @staticmethod
    def convert_to(r, g, b):
        r, g, b = r / 255, g / 255, b / 255
        min_value = min(r, g, b)
        max_value = max(r, g, b)

        l = (min_value + max_value) / 2

        if min_value == max_value:
            return (0, 0, round(l * 100))

        if l <= 0.5:
            s = (max_value - min_value) / (max_value + min_value)
        elif l > 0.5:
            s = (max_value - min_value) / (2.0 - max_value - min_value)

        if max_value == r:
            h = (g - b) / (max_value - min_value)
        elif max_value == g:
            h = 2.0 + (b - g) / (max_value - min_value)
        elif max_value == b:
            h = 4.0 + (r - g) / (max_value - min_value)

        return tuple(round(x) for x in (h * 60, s * 100, l * 100))


class Cmyk(ColorSpace):
    @staticmethod
    def convert_from(c, m, y, k) -> tuple[int, ...]:
        c = c / 100.0
        m = m / 100.0
        y = y / 100.0
        k = k / 100.0

        r = 255.0 - ((min(1.0, c * (1.0 - k) + k)) * 255.0)
        g = 255.0 - ((min(1.0, m * (1.0 - k) + k)) * 255.0)
        b = 255.0 - ((min(1.0, y * (1.0 - k) + k)) * 255.0)

        return tuple(round(x) for x in (r, g, b))

    @staticmethod
    def convert_to(r, g, b) -> tuple[int, ...]:
        if (r, g, b) == (0, 0, 0):
            return (0, 0, 0, 100)

        c, m, y = (1 - x / 255 for x in (r, g, b))

        k = min(c, m, y)
        c = c - k
        m = m - k
        y = y - k

        return tuple(round(x * 100) for x in (c, m, y, k))


def cmyk(c, m, y, k):
    return Hex.convert_to(*Cmyk.convert_from(c, m, y, k))


def hsl(h, s, l):
    return Hex.convert_to(*Hsl.convert_from(h, s, l))


def hsv(h, s, v):
    return Hex.convert_to(*Hsv.convert_from(h, s, v))


def rgb(r, g, b):
    return Hex.convert_to(r, g, b)


@dataclass
class Color:
    red: int
    green: int
    blue: int

    def __init__(self, color: str = None, **kwargs):
        if len(kwargs) != 1:
            if color:
                value = color
                factory = Hex
            else:
                raise ColorError("multiple colors specified")
        else:
            space = tuple(kwargs.keys())[0]
            value = kwargs[space]
            try:
                factory = {
                    "cmyk": Cmyk,
                    "hex": Hex,
                    "hsl": Hsl,
                    "hsv": Hsv,
                    "rgb": Rgb,
                }[space]
            except KeyError:
                raise ColorError(f"invalid color space: {space}") from None

        if factory is Hex:
            value = (value,)

        self.red, self.green, self.blue = factory.convert_from(*value)

    def __to_tcl__(self) -> str:
        return Hex.convert_to(self.red, self.green, self.blue)

    @classmethod
    def __from_tcl__(cls, tcl_value: str) -> Color:
        return Color(rgb=Hex.convert_from(tcl_value))

    @property
    def is_dark(self) -> bool:
        return ((self.red * 299 + self.green * 587 + self.blue * 114) / 1000) < 128

    def invert(self) -> Color:
        return Color(rgb=(255 - self.red, 255 - self.green, 255 - self.blue))

    __invert__ = invert

    def mix(self, other: Color, ratio: float) -> Color:
        if not isinstance(other, Color):
            raise TypeError

        a, b = Fraction.from_float(ratio).as_integer_ratio()
        amount_of_clr_1 = 1 / (a + b) * a
        amount_of_clr_2 = 1 / (a + b) * b

        r, g, b = (
            round(amount_of_clr_1 * value1 + amount_of_clr_2 * value2)
            for value1, value2 in zip(self.rgb, other.rgb)
        )
        return Color(rgb=(r, g, b))

    def __or__(self, other: Color) -> Color:
        return self.mix(other, 1 / 1)

    @property
    def rgb(self) -> tuple[int, int, int]:
        return (self.red, self.green, self.blue)

    @property
    def hex(self) -> str:
        return Hex.convert_to(self.red, self.green, self.blue)

    @property
    def hsl(self) -> tuple[int, int, int]:
        return Hsl.convert_to(self.red, self.green, self.blue)

    @property
    def hsv(self) -> tuple[int, int, int]:
        return Hsv.convert_to(self.red, self.green, self.blue)

    @property
    def cmyk(self) -> tuple[int, int, int, int]:
        return Cmyk.convert_to(self.red, self.green, self.blue)



from __future__ import annotations

from tukaan._base import Container, TkWidget, WidgetBase
from tukaan._props import padding, set_padding


class Frame(WidgetBase, Container):
    _tcl_class = "ttk::frame"

    padding = padding

    def __init__(
        self,
        parent: TkWidget,
        padding: int | tuple[int, ...] | None = None,
        tooltip: str | None = None,
    ) -> None:
        WidgetBase.__init__(self, parent, tooltip=tooltip)

        set_padding(self, padding)



from __future__ import annotations

import contextlib
from collections.abc import Iterator

from tukaan._base import Container, TkWidget, WidgetBase
from tukaan._props import focusable
from tukaan._tcl import Tcl
from tukaan.enums import Orientation
from tukaan.exceptions import TclError

from .frame import Frame


class Pane(Frame):
    _widget: SplitView

    def __init__(
        self,
        *,
        padding: int | tuple[int, ...] | None = None,
        tooltip: str | None = None,
        weight: int | None = None,
    ):
        Frame.__init__(self, self._widget, padding=padding, tooltip=tooltip)

        self._stored_options = {"weight": weight}
        self.append()

    def __repr__(self):
        return f"<tukaan.SplitView.Pane in {self.parent}; tcl_name={self._name}>"

    def append(self) -> None:
        if self in self._widget:
            self.move(-1)
            return None

        Tcl.call(None, self._widget, "add", self, *Tcl.to_tcl_args(**self._stored_options))
        self._widget.panes.append(self)

    def move(self, new_index: int) -> None:
        self._widget.panes.remove(self)
        self._widget.panes.insert(new_index, self)

        if new_index == -1:
            new_index = "end"  # type: ignore

        Tcl.call(None, self._widget, "insert", new_index, self)

    def remove(self) -> None:
        with contextlib.suppress(TclError):
            Tcl.call(None, self._widget, "forget", self)
        self._widget.panes.remove(self)

    @property
    def weight(self) -> int | None:
        if self in self._widget:
            return Tcl.call(int, self._widget, "pane", self, "-weight")
        else:
            return self._stored_options.get("weight", 0)

    @weight.setter
    def weight(self, value: int) -> None:
        if self in self._widget:
            Tcl.call(None, self._widget, "pane", self, "-weight", value)
        self._stored_options["weight"] = value


class SplitView(WidgetBase, Container):
    _tcl_class = "ttk::panedwindow"

    focusable = focusable

    def __init__(
        self,
        parent: TkWidget,
        orientation: Orientation | None = None,
        *,
        focusable: bool | None = None,
        tooltip: str | None = None,
    ) -> None:
        WidgetBase.__init__(self, parent, takefocus=focusable, orient=orientation, tooltip=tooltip)

        self.Pane = Pane
        setattr(self.Pane, "_widget", self)

        self.panes = []
        self._orientation = orientation

    def __len__(self) -> int:
        return len(self.panes)

    def __iter__(self) -> Iterator[Pane]:
        return iter(self.panes)

    def __contains__(self, pane: Pane) -> bool:
        return pane in self.panes

    def __getitem__(self, index: int) -> Pane:
        return self.panes[index]

    def _repr_details(self):
        return f"contains {len(self)} panes"

    def lock_panes(self):
        Tcl.call(None, "bindtags", self, (self, ".", "all"))

    def unlock_panes(self):
        Tcl.call(None, "bindtags", self, (self, "TPanedwindow", ".", "all"))

    @property
    def orientation(self):  # read-only
        return self._orientation



from __future__ import annotations

from typing import Callable, Iterable

from tukaan._base import TkWidget, WidgetBase
from tukaan._props import cget, config
from tukaan._tcl import Tcl
from tukaan.colors import Color

from .textbox import TextBox


class SpinBox(TextBox):
    _tcl_class = "ttk::spinbox"

    def __init__(
        self,
        parent: TkWidget,
        values: Iterable[str | float] | range | None = None,
        *,
        cycle: bool | None = None,
        fg_color: str | Color | None = None,
        focusable: bool | None = None,
        hide_chars: bool | None = False,
        hide_chars_with: str | None = "•",
        on_select: Callable[[str], None] | None = None,
        step: int | None = None,
        text_align: str | None = None,
        tooltip: str | None = None,
        user_edit: bool = True,
        value: str | float | None = None,
        width: int | None = None,
    ) -> None:
        self._prev_show_char = hide_chars_with
        if not hide_chars:
            hide_chars_with = None

        self._original_cmd = on_select
        if on_select is not None:
            func = on_select
            on_select = lambda: func(self.get())  # type: ignore

        WidgetBase.__init__(
            self,
            parent,
            command=on_select,
            foreground=fg_color,
            justify=text_align,
            show=hide_chars_with,
            state=None if user_edit else "readonly",
            takefocus=focusable,
            tooltip=tooltip,
            width=width,
            wrap=cycle,
        )

        self._set_values(values, step)

        if value is not None:
            self.set(value)
        elif values:
            self.set(values[0])
        else:
            self.set("0")

        self.bind("<<Increment>>", f"+{self._name} selection clear")
        self.bind("<<Decrement>>", f"+{self._name} selection clear")
        self.bind("<FocusOut>", f"+{self._name} selection clear")

    def _set_values(self, values, arg_step) -> None:
        if isinstance(values, range):
            start, stop, step = values.start, values.stop, values.step
            if arg_step:
                step = arg_step

            return Tcl.call(
                None,
                self,
                "configure",
                *Tcl.to_tcl_args(from_=start, to=stop - step, increment=step),
            )

        Tcl.call(None, self, "configure", *Tcl.to_tcl_args(values=values))

    def set(self, value: str | float | None) -> None:
        Tcl.call(None, self, "set", value)

    value = property(TextBox.get, set)

    @property
    def values(self) -> list[str] | range:
        result = Tcl.call([str], self, "cget", "-values")

        if not result:
            start = Tcl.call(float, self, "cget", "-from") or 0
            stop = Tcl.call(float, self, "cget", "-to") or 0
            step = Tcl.call(float, self, "cget", "-increment")
            if isinstance(step, float):
                step = 1
            return range(start, stop, step)

        return result

    @values.setter
    def values(self, values: Iterable[str | float] | range | None) -> None:
        self._set_values(values, None)

    @property
    def cycle(self) -> bool:
        return cget(self, bool, "-wrap")

    @cycle.setter
    def cycle(self, value: bool) -> None:
        config(self, wrap=value)

    @property
    def step(self) -> float:
        return cget(self, float, "-increment")

    @step.setter
    def step(self, value: float) -> None:
        config(self, increment=value)

    @property
    def on_select(self) -> Callable[[str], None] | None:
        return self._original_cmd

    @on_select.setter
    def on_select(self, func: Callable[[str], None] | None) -> None:
        self._original_cmd = func
        if func is not None:
            value = lambda: func(self.get())
        else:
            value = ""
        config(self, command=value)



from __future__ import annotations

import ctypes
import re
import sys
from abc import ABC, abstractmethod
from typing import Any, Callable, Final

from ._data import Size
from ._tcl import Tcl
from .colors import Color
from .enums import BackdropEffect, WindowState
from .exceptions import TclError

DWMWA_TRANSITIONS_FORCEDISABLED: Final[int] = 3
DWMWA_NONCLIENT_RTL_LAYOUT: Final[int] = 6
DWMWA_FORCE_ICONIC_REPRESENTATION: Final[int] = 7
DWMWA_USE_IMMERSIVE_DARK_MODE: Final[int] = 20
DWMWA_BORDER_COLOR: Final[int] = 34
DWMWA_CAPTION_COLOR: Final[int] = 35
DWMWA_TEXT_COLOR: Final[int] = 36

# https://github.com/minusium/MicaForEveryone/blob/master/MicaForEveryone.Win32/DesktopWindowManager.cs#L15
DWMWA_SYSTEMBACKDROP_TYPE: Final[int] = 38

# https://github.com/Brouilles/win32-mica/blob/main/main.cpp#L119
DWMWA_MICA_EFFECT: Final[int] = 1029


class AccentPolicy(ctypes.Structure):
    _fields_ = [
        ("AccentState", ctypes.c_int),
        ("AccentFlags", ctypes.c_int),
        ("GradientColor", ctypes.c_uint),
        ("AnimationId", ctypes.c_int),
    ]


class WindowCompositionAttributeData(ctypes.Structure):
    _fields_ = [
        ("Attribute", ctypes.c_int),
        ("Data", ctypes.POINTER(ctypes.c_int)),
        ("SizeOfData", ctypes.c_size_t),
    ]


class WMProperties:
    def _get_title(self) -> str:
        return Tcl.call(str, "wm", "title", self._wm_path)

    def _set_title(self, new_title: str) -> None:
        Tcl.call(None, "wm", "title", self._wm_path, new_title)

    title = property(_get_title, _set_title)

    @Tcl.update_before
    def _get_x(self) -> int:
        return Tcl.call(int, "winfo", "x", self._wm_path)

    @Tcl.update_after
    def _set_x(self, new_x: int) -> None:
        Tcl.call(None, "wm", "geometry", self._wm_path, f"+{new_x}+{self._get_y()}")

    x = property(_get_x, _set_x)

    @Tcl.update_before
    def _get_y(self) -> int:
        return Tcl.call(int, "winfo", "y", self._wm_path)

    @Tcl.update_after
    def _set_y(self, new_y: int) -> None:
        Tcl.call(None, "wm", "geometry", self._wm_path, f"+{self._get_x()}+{new_y}")

    y = property(_get_y, _set_y)

    @Tcl.update_before
    def _get_width(self) -> int:
        return Tcl.call(int, "winfo", "width", self._wm_path)

    @Tcl.update_after
    def _set_width(self, new_width: int) -> None:
        Tcl.call(None, "wm", "geometry", self._wm_path, f"{new_width}x{self._get_height()}")

    width = property(_get_width, _set_width)

    @Tcl.update_before
    def _get_height(self) -> int:
        return Tcl.call(int, "winfo", "height", self._wm_path)

    @Tcl.update_after
    def _set_height(self, new_height: int) -> None:
        Tcl.call(None, "wm", "geometry", self._wm_path, f"{self._get_width()}x{new_height}")

    height = property(_get_height, _set_height)

    @Tcl.update_before
    def _get_size(self) -> Size:
        return Size(
            *map(
                int,
                re.split(r"x|\+", Tcl.call(str, "wm", "geometry", self._wm_path))[:2],
            )
        )

    @Tcl.update_after
    def _set_size(self, new_size: Size | tuple[int, int] | int) -> None:
        if isinstance(new_size, int):
            new_size = (new_size,) * 2

        Tcl.call(None, "wm", "geometry", self._wm_path, "{}x{}".format(*new_size))

    size = property(_get_size, _set_size)


class WindowManagerBase(ABC, WMProperties):
    _name: str
    _wm_path: str
    generate_event: Callable[[str], None]

    _current_state = WindowState.Normal

    def _generate_state_event(self):
        prev_state = self._current_state
        new_state = self.state

        if new_state == prev_state:
            return

        event = None
        un_event = None

        if new_state == "normal":
            event = "<<Restore>>"
        elif new_state == "maximized":
            event = "<<Maximize>>"
        elif new_state == "minimized":
            event = "<<Minimize>>"
        elif new_state == "hidden":
            event = "<<Hide>>"
        elif new_state == "fullscreen":
            event = "<<Fullscreen>>"

        if prev_state == "maximized":
            un_event = "<<Unmaximize>>"
        elif prev_state == "minimized":
            un_event = "<<Unminimize>>"
        elif prev_state == "hidden":
            un_event = "<<Unhide>>"
        elif prev_state == "fullscreen":
            un_event = "<<Unfullscreen>>"

        self.generate_event("<<StateChanged>>")

        if event:
            self.generate_event(event)

        if un_event:
            self.generate_event(un_event)

        self._current_state = new_state

    def minimize(self) -> None:
        Tcl.call(None, "wm", "iconify", self._wm_path)

    @abstractmethod
    def maximize(self) -> None:
        pass

    def fullscreen(self) -> None:
        Tcl.call(None, "wm", "attributes", self._wm_path, "-fullscreen", True)

    def focus(self) -> None:
        Tcl.call(None, "focus", "-force", self._wm_path)

    def group(self, other: WindowManagerBase) -> None:
        Tcl.call(None, "wm", "group", self._wm_path, other._wm_path)

    def hide(self):
        Tcl.call(None, "wm", "withdraw", self._wm_path)

    def unhide(self):
        Tcl.call(None, "wm", "deiconify", self._wm_path)

    def on_close(self, func: Callable[[WindowManagerBase], None]) -> Callable[[], None]:
        def wrapper() -> None:
            if func(self):
                self.destroy()

        Tcl.call(None, "wm", "protocol", self._wm_path, "WM_DELETE_WINDOW", wrapper)
        return wrapper

    @property
    def id(self) -> int:
        return int(Tcl.call(str, "winfo", "id", self._wm_path), 16)

    @property
    def is_focused(self) -> bool:
        return Tcl.call(str, "focus", "-displayof") == self._wm_path

    @property
    def state(self):
        try:
            state = Tcl.call(str, "wm", "state", self._wm_path)
        except TclError as e:
            if not Tcl.call(bool, "winfo", "exists", self._wm_path):
                return WindowState.Closed
            raise e

        if state == "normal" and Tcl.windowing_system != "x11":  # needs further checking on X11
            return WindowState.Normal
        elif state == "iconic":
            return WindowState.Minimized
        elif state == "zoomed":
            return WindowState.Maximized
        elif state == "withdrawn":
            return WindowState.Hidden
        # fmt: off
        elif (
            Tcl.windowing_system == "x11" and
            Tcl.call(bool, "wm", "attributes", self._wm_path, "-zoomed")
        ):
            return WindowState.Maximized
        # fmt: on
        elif Tcl.call(bool, "wm", "attributes", self._wm_path, "-fullscreen"):
            return WindowState.FullScreen

        return WindowState.Normal

    ### Windows only stuff ###
    def set_backdrop_effect(self, *args, **kwargs) -> None:
        pass

    @property
    def is_toolwindow(self) -> None:
        pass

    @is_toolwindow.setter
    def is_toolwindow(self, *args) -> None:
        pass

    @property
    def border_color(self) -> None:
        pass

    @border_color.setter
    def border_color(self, *args) -> None:
        pass

    @property
    def immersive_dark_mode(self) -> None:
        pass

    @immersive_dark_mode.setter
    def immersive_dark_mode(self, *args) -> None:
        pass


class DWM(WindowManagerBase):
    """Windows"""

    def maximize(self) -> None:
        Tcl.call(None, "wm", "state", self._wm_path, "zoomed")

    @property
    def is_toolwindow(self) -> bool:
        return Tcl.call(bool, "wm", "attributes", self._wm_path, "-toolwindow")

    @is_toolwindow.setter
    def is_toolwindow(self, value: bool) -> None:
        Tcl.call(None, "wm", "attributes", self._wm_path, "-toolwindow", value)

    @property
    def border_color(self) -> Color:
        ...

    @border_color.setter
    def border_color(self, value: Color | str) -> None:
        if isinstance(value, Color):
            value = value.hex

        self._dwm_set_window_attribute(DWMWA_BORDER_COLOR, self._get_bgr_color(value))

    def set_backdrop_effect(
        self,
        effect: BackdropEffect = BackdropEffect.Blur,
        tint: Color | str | None = None,
        tint_opacity: float = 0.2,
    ) -> None:
        # Useful links:
        # https://github.com/Peticali/PythonBlurBehind/blob/main/blurWindow/blurWindow.py
        # https://github.com/sourcechord/FluentWPF/blob/master/FluentWPF/Utility/AcrylicHelper.cs

        build = sys.getwindowsversion().build
        has_mica = build >= 22000
        has_backdrop_type = build >= 22523

        if not has_mica and effect is BackdropEffect.Mica:
            # Mica is available only on Windows 11 build 22000+, fall back to acrylic
            effect = BackdropEffect.Acrylic

        if effect is BackdropEffect.No:
            gradient_color = 0

            # Remove `-transparentcolor`
            Tcl.eval(None, f"wm attributes {self._wm_path} -transparentcolor {{}}")

            self.unbind("<<Unmaximize>>")
        else:
            # Make an ABGR color from `tint` and `tint_opacity`
            # If `tint` is None, use the window background color
            bg_color = Tcl.eval(str, "ttk::style lookup . -background")

            if isinstance(tint, Color):
                tint_color = tint.hex
            else:
                if tint is None:
                    tint = Tcl.eval((int,), f"winfo rgb . {bg_color}")
                tint_color = Color(rgb=tuple(value >> 8 for value in tint)).hex

            tint_hex_a = str(hex(int(255 * tint_opacity)))[2:]
            tint_hex_bgr = self._get_bgr_color(tint_color)
            gradient_color = int(tint_hex_a + tint_hex_bgr, 16)

            # Make the window background, and each pixel with the same color
            # as the window background fully transparent.
            # These are the areas that will be blurred.
            Tcl.eval(None, f"wm attributes {self._wm_path} -transparentcolor {bg_color}")

            # When the window has `transparentcolor`, the whole window becomes unusable after unmaximizing.
            # Therefore we bind it to the <<Unmaximize>> event (see it below: WindowManager._generate_state_event),
            # so every time it changes state, it calls the _fullredraw method.
            # See DWM._fullredraw.__doc__ for more.
            self.bind("<<Unmaximize>>", self._fullredraw)

        accent_policy = AccentPolicy()
        accent_policy.AccentFlags = 2  # ACCENT_ENABLE_BLURBEHIND  <- try this with '4' bruhh :D
        accent_policy.AccentState = effect.value
        accent_policy.GradientColor = gradient_color

        wcad = WindowCompositionAttributeData()
        wcad.Attribute = 19  # WCA_ACCENT_POLICY
        wcad.Data = ctypes.cast(ctypes.pointer(accent_policy), ctypes.POINTER(ctypes.c_int))
        wcad.SizeOfData = ctypes.sizeof(accent_policy)

        self._set_window_composition_attribute(wcad)

        if effect is BackdropEffect.Mica:
            if has_backdrop_type:
                # https://github.com/minusium/MicaForEveryone/blob/master/MicaForEveryone.Win32/PInvoke/DWM_SYSTEMBACKDROP_TYPE.cs
                # Mica effect is 2
                self._dwm_set_window_attribute(DWMWA_SYSTEMBACKDROP_TYPE, 2)
            elif has_mica:
                self._dwm_set_window_attribute(DWMWA_MICA_EFFECT, 1)
        elif effect is BackdropEffect.No:
            # Remove Mica
            if has_backdrop_type:
                # None effect is 1
                self._dwm_set_window_attribute(DWMWA_SYSTEMBACKDROP_TYPE, 1)
            elif has_mica:
                self._dwm_set_window_attribute(DWMWA_MICA_EFFECT, 0)

    def _fullredraw(self) -> None:
        """
        Internal method

        Neither user32.RedrawWindow nor user32.UpdateWindow redraws the titlebar,
        so we need to explicitly minimize and then restore the window.
        To avoid visual effects, disable transitions on the window for that time.
        """

        self._dwm_set_window_attribute(DWMWA_TRANSITIONS_FORCEDISABLED, 1)
        self.minimize()
        self.restore()
        self._dwm_set_window_attribute(DWMWA_TRANSITIONS_FORCEDISABLED, 0)

    def _get_WHND(self):
        return int(Tcl.call(str, "wm", "frame", self._wm_path), 16)

    def _get_bgr_color(self, color: str) -> str:
        return color[5:7] + color[3:5] + color[1:3]

    def _dwm_set_window_attribute(self, rendering_policy: int, value: Any) -> None:
        value = ctypes.c_int(value)

        ctypes.windll.dwmapi.DwmSetWindowAttribute(
            self._get_WHND(),
            rendering_policy,
            ctypes.byref(value),
            ctypes.sizeof(value),
        )

    def _set_window_composition_attribute(self, wcad: WindowCompositionAttributeData) -> None:
        ctypes.windll.user32.SetWindowCompositionAttribute(self._get_WHND(), wcad)


class Quartz(WindowManagerBase):
    """macOS"""

    def maximize(self) -> None:
        Tcl.call(None, "wm", "attributes", self._wm_path, "-zoomed", True)


class X11(WindowManagerBase):
    """GNU/Linux"""

    def maximize(self) -> None:
        Tcl.call(None, "wm", "attributes", self._wm_path, "-zoomed", True)


if sys.platform == "win32":
    WindowManager = DWM
elif sys.platform == "darwin":
    WindowManager = Quartz
else:
    WindowManager = X11



from setuptools import setup
from tukaan import __author__, __version__


def get_requirements():
    with open("requirements.txt", "r") as file:
        for line in [x.strip() for x in file]:
            if line and not line.startswith("#"):
                yield line


with open("README.md", "r") as file:
    long_description = file.read()


setup(
    name="tukaan",
    version=__version__,
    author=__author__,
    author_email="rdbende@gmail.com",
    url="https://github.com/tukaan/tukaan",
    license="MIT license",
    python_requires=">=3.7",
    install_requires=list(get_requirements()),
    description="A modern, cross platform Python toolkit for creating desktop GUI applications, based on Tcl/Tk.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["gui", "ui", "tukaan", "tkinter", "ttk", "tcl", "tk", "tcl/tk"],
    packages=[
        "tukaan",
        "tukaan/Serif",
        "tukaan/themes",
        "tukaan/tkdnd",
        "tukaan/widgets",
    ],
    package_data={
        "": [
            "Serif/pkg/*.*",
            "tkdnd/*/*.*",
        ]
    },
    include_package_data=True,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Tcl",
        "Programming Language :: C",
        "Programming Language :: C++",
        "Typing :: Typed",
    ],
)



import sys
from pathlib import Path

from tukaan._tcl import Tcl

os = {"linux": "linux", "darwin": "mac", "win32": "windows"}[sys.platform]


def load() -> None:
    system = os

    system += "_x64" if sys.maxsize > 2**32 else "_x32"
    ext = Tcl.call(str, "info", "sharedlibextension")

    libfilename = "serif_" + system + ext
    Tcl.call(None, "load", Path(__file__).parent / "pkg" / libfilename, "Serif")



from __future__ import annotations

from typing import Callable

from tukaan._base import InputControl, TkWidget, WidgetBase
from tukaan._props import config, focusable, link, text, width
from tukaan._tcl import Tcl
from tukaan._variables import Boolean


class CheckBox(WidgetBase, InputControl):
    _tcl_class = "ttk::checkbutton"

    _variable: Boolean

    focusable = focusable
    link = link
    text = text
    width = width

    def __init__(
        self,
        parent: TkWidget,
        text: str,
        *,
        focusable: bool | None = None,
        link: Boolean | None = None,
        on_click: Callable[[bool], None] | None = None,
        tooltip: str | None = None,
        value: bool = False,
        width: int | None = None,
    ) -> None:

        if link is None:
            self._variable = link = Boolean(value)
        else:
            self._variable = link

        self._original_cmd = on_click
        if on_click is not None:
            func = on_click
            on_click = lambda: func(self._variable.get())  # type: ignore

        WidgetBase.__init__(
            self,
            parent,
            command=on_click,
            offvalue=False,
            onvalue=True,
            takefocus=focusable,
            text=text,
            tooltip=tooltip,
            variable=link,
            width=width,
        )

    def invoke(self) -> None:
        """Invokes the checkbox, as if it were clicked"""

        Tcl.call(None, self, "invoke")

    def select(self) -> bool:
        """Selects the checkbox"""

        return self._variable.set(True)

    def deselect(self) -> bool:
        """Deselects the checkbox"""

        return self._variable.set(False)

    def toggle(self) -> bool:
        """Toggles the state of the checkbox"""

        return ~self._variable.set()

    @property
    def value(self) -> bool:
        return self._variable.get()

    @value.setter
    def value(self, value: bool) -> None:
        self._variable.set(value)

    selected = value

    @property
    def on_click(self) -> Callable[[bool], None] | None:
        return self._original_cmd

    @on_click.setter
    def on_click(self, func: Callable[[bool], None] | None) -> None:
        self._original_cmd = func
        if func is not None:
            value = lambda: func(self._variable.get())
        else:
            value = ""
        config(self, command=value)
